#include "jsapi.h"
#include "jscntxt.h"    //Used for debugging purposes with runtime and context.

#include <vector>
#include <map>
#include <string>

#include "./private/primpl.h"   //Added this so we could debug into PR stuctures!

#ifdef WIN32

   #pragma warning(disable : 4786)

   #include <process.h>
   #include <windows.h>

   #if _MSC_VER >= 1300
   #include <Tlhelp32.h>
   #endif

   #include "StackWalker.h"

  // Simple implementation of an additional output to the console:
  class MyStackWalker : public StackWalker  
  {
  public:
    MyStackWalker() : StackWalker() {}
    MyStackWalker(DWORD dwProcessId, HANDLE hProcess) : StackWalker(dwProcessId, hProcess) {}
    virtual void OnOutput(LPCSTR szText) { printf(szText); StackWalker::OnOutput(szText); }
  };

  CRITICAL_SECTION*       g_pContextPoolLock;
  typedef std::map<HANDLE, unsigned long> HANDLEMAP;
  HANDLEMAP               g_oThreadHandleMap;    /* MAP containing list of HANDLES to all running child threads*/

  #include <conio.h>  //used for _kbhit function

  #define RSLEEP(iMilliSecs) Sleep(##iMilliSecs)
  #define LockPool  EnterCriticalSection(g_pContextPoolLock);
  #define UnlockPool  LeaveCriticalSection(g_pContextPoolLock);

#else /*Linux and others */

  #include "pthread.h"
  pthread_mutex_t*        g_pContextPoolLock;

  #define RSLEEP(iMilliSecs) usleep(##iMilliSecs*10)

  #define LockPool  pthread_mutex_lock(g_pContextPoolLock);
  #define UnlockPool  pthread_mutex_unlock(g_pContextPoolLock);

/**
 Linux (POSIX) implementation of _kbhit().
 Morgan McGuire, morgan@cs.brown.edu
 */
#include <stdio.h>
#include <sys/select.h>
#include <termios.h>
#include <stropts.h>

int _kbhit() {
    static const int STDIN = 0;
    static bool initialized = false;

    if (! initialized) {
        // Use termios to turn off line buffering
        termios term;
        tcgetattr(STDIN, &term);
        term.c_lflag &= ~ICANON;
        tcsetattr(STDIN, TCSANOW, &term);
        setbuf(stdin, NULL);
        initialized = true;
    }

    int bytesWaiting;
    ioctl(STDIN, FIONREAD, &bytesWaiting);
    return bytesWaiting;
}
#endif


JSClass global_class = {
    "global", JSCLASS_HAS_PRIVATE,
    JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,
    JS_EnumerateStub,JS_ResolveStub,JS_ConvertStub,JS_FinalizeStub
};

//Pool of JS Contexts we use for executing server side scripts.
typedef std::vector<JSContext*> JSCONTEXT_POOL_LIST;

//A pool of contexts that are free and can be used to execute a script.
JSCONTEXT_POOL_LIST     g_oFreeContextsList;
JSRuntime*              g_pRuntime;
JSScript*               g_pJSCompiledScript;
JSObject*               g_pJSScriptObj;
std::string             g_sCompiledScriptSource;
std::string             g_sScriptName;
long                    g_liNumActiveThreads;
long                    g_liNumThreadsStarting;
long                    g_liTotalExecutions;
long                    g_liTotalExecutionsGC;



JSContext* GetFreeContextFromPool()
{
    JSContext* pFreeContext=NULL;
    
    //Pop the first context from the top if any in the free pool.    
    LockPool
    if(g_oFreeContextsList.size() > 0)
    {
        pFreeContext = g_oFreeContextsList.front();
        g_oFreeContextsList.erase(g_oFreeContextsList.begin());
    }
    UnlockPool

    //If none are free then create a new one.
    if(!pFreeContext)
    {
        pFreeContext = JS_NewContext(g_pRuntime, 1024*1024*2);    //2 megs to start.
        if(pFreeContext)
        {
	        JS_SetContextThread(pFreeContext);  //Associate the context with our thread.
            JS_BeginRequest(pFreeContext);  //Must start a request before allocating new objects!

            JSObject*  pGlobal=NULL;
            pGlobal = JS_NewObject(pFreeContext, &global_class, NULL, NULL);
            JS_SetGlobalObject(pFreeContext, pGlobal);
        }
    }
    else
    {
        JS_SetContextThread(pFreeContext);  //Associate the context with our thread.
        JS_BeginRequest(pFreeContext);  //Must start a request before allocating new objects!
    }

    return pFreeContext;
}

void ReturnContextToPool(JSContext* pContext)
{
    JS_ClearContextThread(pContext);  //Clear the thread associated with the context

    LockPool
    --g_liNumActiveThreads;
    ++g_liTotalExecutions;  //How many times have we run it?
    g_oFreeContextsList.push_back(pContext);

    if(g_liNumActiveThreads >= 0)
        printf("Pool Size(%d)  Active Threads(%d) Total Executions(%d)\n", g_oFreeContextsList.size(), g_liNumActiveThreads, g_liTotalExecutions);

    UnlockPool
}

long 
GetNumThreadsStarting()
{
    return g_liNumThreadsStarting;
}

long
GetNumActiveThreads()
{
    return g_liNumActiveThreads;
}

void
IncrementNumThreadsStarting()
{
#ifdef WIN32
    InterlockedIncrement(&g_liNumThreadsStarting);
#else 
    //Anybody know how to do atomic increment in Linux??  If so fix here...
    LockPool
    ++g_liNumThreadsStarting;
    UnlockPool;
#endif
}

long
IncrementNumActiveThreads()
{
    long liNumActive=0;
    
#ifdef WIN32
    liNumActive = InterlockedIncrement(&g_liNumActiveThreads);
    InterlockedDecrement(&g_liNumThreadsStarting);
#else
    //Anybody know how to do atomic increment in Linux??  If so fix here...
    LockPool
    liNumActive = ++g_liNumActiveThreads;
    ++g_liNumThreadsStarting
    UnlockPool;
#endif

    return liNumActive;
}

long
GetTotalThreadCount()
{
    long liTotalThreads=0;
    liTotalThreads = g_liNumActiveThreads + g_liNumThreadsStarting;
    return liTotalThreads;
}

#ifdef WIN32
void WorkerThreadProc(void* /*pNothing*/)
#else
void* WorkerThreadProc(void* /*pNothing*/)
#endif
{
    printf("%d Started new thread. Active thread count(%d)\n", GetCurrentThreadId(), IncrementNumActiveThreads());

#ifdef WIN32
    //We keep a list of thread handles for WIN32
    HANDLE myHandle=NULL;
    LockPool
    //Create a thread handle that can be used by other threads from the outside.
    if(DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), GetCurrentProcess(), &myHandle, 0, FALSE, DUPLICATE_SAME_ACCESS ))
        g_oThreadHandleMap.insert(HANDLEMAP::value_type(myHandle, GetCurrentThreadId()));
    UnlockPool
#endif

    JSContext* pContext = GetFreeContextFromPool();

    //Initialize standard classes.
    JSObject* pGlobal = JS_GetGlobalObject(pContext);
    JSBool bOK = JS_FALSE;
    bOK = JS_InitStandardClasses(pContext, pGlobal);

 	//Next line is for debugging a possible problem with GC crash. TODO Remove this.
	//unsigned int savedGCNumber = 0;
	//savedGCNumber = g_pRuntime->gcNumber;
  
    if(NULL == g_pJSCompiledScript)
    {
        //local copies in case two threads to this at the same time.
        JSScript* pJSCompiledScript;
        pJSCompiledScript = JS_CompileScript(pContext, pGlobal, g_sCompiledScriptSource.c_str(), g_sCompiledScriptSource.size(), g_sScriptName.c_str(), 1);
        if(pJSCompiledScript)
        {
            //Save the compiled version in the runtime for faster execution.
            //Root it to protect from GC.
            LockPool
            if(NULL == g_pJSCompiledScript)
            {
                g_pJSCompiledScript = pJSCompiledScript;
                UnlockPool
                g_pJSScriptObj = JS_NewScriptObject(pContext, g_pJSCompiledScript);
                JSBool bOK=0;
                bOK = JS_AddNamedRootRT(g_pRuntime, &(g_pJSScriptObj), g_sScriptName.c_str());
            }
            else
            {
                UnlockPool
            }
        }
    }

    //Run the script!
    jsval oReturnVal;
    bOK = JS_ExecuteScript(pContext, pGlobal, g_pJSCompiledScript, &oReturnVal);
    
    //Don't destroy the context.  Just clean it up and put it back in the pool of contexts
	//so it can be re-used.
	//JS_SetErrorReporter(pContext, pSavedErrorReporter);  //Restore any previous error reporter.

    JS_ClearNewbornRoots(pContext);
	JS_ClearRegExpStatics(pContext);
	JS_ClearPendingException(pContext);
	JS_ClearScope(pContext, pGlobal);                    //Clear the global
    JS_EndRequest(pContext);

 	//To a full Garbage collection sweep every 50 hits
	//otherwise just maybe GC.
	if(g_liTotalExecutions > 0
	&& g_liTotalExecutions%50 == 0
    && g_liTotalExecutionsGC != g_liTotalExecutions)
	{
        g_liTotalExecutionsGC = g_liTotalExecutions;
   		printf("%d STARTING GARBAGE COLLECTION! Hit count(%ld)\n", GetCurrentThreadId(), g_liTotalExecutions);
		JS_GC(pContext);
	}
	else
    {
        /* No More MAYBE Stuff. Force GC much more often to cause crashes quicker */
  		printf("%d FORCE GC... Hit count(%ld)\n", GetCurrentThreadId(), g_liTotalExecutions);
        JS_GC(pContext);

        //printf("%d MAYBE GC... Hit count(%ld)\n", GetCurrentThreadId(), g_liTotalExecutions);
        //JS_MaybeGC(pContext);
    }

	ReturnContextToPool(pContext);

    RSLEEP(100);    //Sleep a little to cause crashes..

#ifdef WIN32
    LockPool
    //Remove the handle from the vector
    HANDLEMAP::iterator hIter;
    hIter = g_oThreadHandleMap.find(myHandle);
    if(hIter != g_oThreadHandleMap.end())
        g_oThreadHandleMap.erase(hIter);
    UnlockPool
#endif

}

int main(int /*argc*/, char* /*argv[]*/)
{
   //Setup syncronization objects
#ifdef WIN32
    CRITICAL_SECTION oMutex;
    g_pContextPoolLock = &oMutex;
    InitializeCriticalSection(g_pContextPoolLock);
#else
    pthread_mutex_t oMutex;
    g_pContextPoolLock = &oMutex;
    pthread_mutex_init(g_pContextPoolLock, NULL);
#endif

/* ugly script to work the engine a bit...*/

g_sCompiledScriptSource = " var o =" \
"  {" \
"    total : 0," \
"    scoped : function(obj, fOrName)" \
"    {" \
"      var f = fOrName;" \
"      if(typeof(fOrName) == \"string\")" \
"        f = obj[fOrName];" \
" 	    return function(){ return f.apply(obj, arguments);}" \
"    }," \
"    i:0," \
"    lastn : null," \
"    functiondef :     /* new Function('x', 'return arguments[0]+1;]) */ " \
"    [\"\\u006e\",\"\\u0065\",\"\\u0077\",\"\\u0020\",\"\\u0046\",\"\\u0075\",\"\\u006e\",\"\\u0063\",\"\\u0074\",\"\\u0069\",\"\\u006f\",\"\\u006e\",\"\\u0028\",\"\\u0027\"," \
"     \"\\u0078\",\"\\u0027\",\"\\u002c\",\"\\u0020\",\"\\u0027\",\"\\u0072\",\"\\u0065\",\"\\u0074\",\"\\u0075\",\"\\u0072\",\"\\u006e\",\"\\u0020\",\"\\u0061\",\"\\u0072\"," \
"     \"\\u0067\",\"\\u0075\",\"\\u006d\",\"\\u0065\",\"\\u006e\",\"\\u0074\",\"\\u0073\",\"\\u005b\",\"\\u0030\",\"\\u005d\",\"\\u002b\",\"\\u0031\",\"\\u003b\",\"\\u0027\",\"\\u0029\"]" \
"  };" \
"  /* findex is used to build the chain of functions.  each string represents a function name, each number is added to the total.  arrays & objects are recurrsed into */" \
"  var findex = [\"a1\", \"a2\", \"a3\", \"a4\", 10, \"a5\", [\"b1\", \"b2\", [\"c1\", \"c2\", \"c3\", \"c4\", \"c5\", [\"d1\", \"d2\", [\"e1\", [\"f1\", \"f2\"]], \"d3\"]], \"b3\"], \"a6\", \"a7\", \"a8\", \"a9\", { x:5, y:10, z:[6, 7, 8] }, \"a10\", \"a11\", 2];" \
"  var a = [\"A\", \"B\", \"C\", \"D\"]; " \
" for(var n in a)" \
"   var q = \"n:\" + n + \" a[n]:\" + a[n];" \
" var x = 3;" \
" if(x==undefined)" \
"   var z = \"NO\";" \
" var z = \"YES\";" \
"  (function(args)  /* anonymous function w/ applied scope */" \
"  {" \
"    var arr = args?args:findex;" \
"    for(var n in arr)  /* iteration/enumeration */" \
"    {" \
"      if(typeof(arr[n])==\"string\") /* typeof literal*/" \
"      {" \
"        this[arr[n]] = eval(\"function() { this.total += \" + this.i + \";\" + ((this.lastn)?(\"this['\" + this.lastn + \"']();\"):\"\") +    \"}\");   /* eval'ed function & closure */" \
"        this.i = eval( (function() { var s=\"\"; for(var i=0;i<this.functiondef.length;i++) { s+=this.functiondef[i]; } return s;}).apply(this) )(this.i);  /* build the eval string using a for loop which generates a new Function call, call that function and apply scope */" \
"        this.lastn = arr[n];" \
"        this.run=this.scoped(this, this[arr[n]]);  /* assign function, scoping closure */" \
"      }" \
"      else if(typeof(arr[n])==\"object\" && arr[n] instanceof Array)  /* typeof & instanceof  */" \
"        arguments.callee.call(this, arr[n]); /* recursion w/ scope */" \
"      else if(typeof(arr[n])==\"number\")" \
"        this.total+=arr[n];" \
"      else if(typeof(arr[n])==\"object\")" \
"        arguments.callee.call(this, arr[n]); /* recursion w/ scope */" \
"    }      " \
"  }).apply(o);" \
"  o.run();";

    g_sScriptName = "CorruptScript.js";
    g_pRuntime = JS_NewRuntime(1024*1024*50);   //50 megs

    g_liNumActiveThreads=0;
    g_liNumThreadsStarting=0;
    g_liTotalExecutions=0;
    bool bWalkedThreadStacks=false;
    unsigned long ulStartTick = 0();
    unsigned long ulStopTick =0;
    bool bAllThreadsRunning=false;
    bool bForever=true;

    while(bForever)
	{
        //Break out whenever the keyboard is hit.
        if(_kbhit())
        {
            break;
        }
        else
        {
            //Start another thread to execute the script yet again!
            //Keep at least X threads going to beat up the engine.
            if(GetTotalThreadCount() < 5)
            {
                IncrementNumThreadsStarting();    //Each thread will decrement when it dies.
#ifdef WIN32
                HANDLE hThread = NULL;
                hThread= (HANDLE) _beginthread(WorkerThreadProc, 0, (void*)NULL);
#else
                int iError=0;
                pthread_t tid;
                iError = pthread_create(&tid, NULL, WorkerThreadProc, NULL);
#endif
                bAllThreadsRunning=false;
            }
            else
            {
                RSLEEP(1); //Sleep for a tiny bit so as not to kill the processor(s)
                if(!bAllThreadsRunning)
                {
#ifdef WIN32
                    ulStartTick = GetTickCount();
#endif
                    bAllThreadsRunning=true;
                    //break;  //Break out for now to cause crashes.
                    
                }
                else
                {
#ifdef WIN32
                    //If all the threads are running and we havn't started any new ones in at least 8
                    //seconds it generally means we are deadlocked!
                    //Dump a complete stack trace for each active thread.
                    ulStopTick = GetTickCount();
                    if(!bWalkedThreadStacks && (ulStopTick-ulStartTick)/1000.0 > 8)
                    {
                        bWalkedThreadStacks=true;
                        printf("\n---------------------------------------------------------------------------\n");
                        printf("Detected Thread Deadlock.  Dumping Stack for each Thread.  Main Thread(%d)\n", GetCurrentThreadId());
                        printf("-----------------------------------------------------------------------------\n");

                        LockPool
                        
                        //Dump out some lock information about our runtime to aid in debugging.
                        printf("JS RUNTIME MEMORY ADDRESSES...\n");
                        printf("rt->gcDone           has Address(%X)\n", g_pRuntime->gcDone);
                        printf("rt->requestDone      has Address(%X)\n", g_pRuntime->requestDone);
                        printf("rt->stateChange      has Address(%X)\n", g_pRuntime->stateChange);
                        printf("rt->atomState->lock  has Address(%X)\n", g_pRuntime->atomState);
                        printf("rt->gcLock           has Address(%X)\n", g_pRuntime->gcLock);

                        MyStackWalker sw;
                        sw.LoadModules();   //Initialize and load up symbols.

                        //Remove the handle from the vector
                        HANDLEMAP::iterator hIter;
                        hIter = g_oThreadHandleMap.begin();
                        while(hIter != g_oThreadHandleMap.end())
                        {
                            printf("\n-----------------------------------\n");
                            printf("Dumping Stack Trace for Thread(%ld)\n", hIter->second);
                            printf("-----------------------------------\n");

                            sw.ShowCallstack(hIter->first);
                            ++hIter;
                        }

                        UnlockPool
                    }
                    else
                    {
                        RSLEEP(5);
                    }
#else
                    RSLEEP(5);  //No stack dumping in linux. Just sleep and start more threads!
#endif
                }
            }
        }
    }


    if(g_pRuntime)
    {
        long liActiveThreads = GetTotalThreadCount();
        long liImpatienceCounter=0;
        if(liActiveThreads > 0)
        {
            //Wait for all active contexts that are executing scripts to terminate.
            //After 1 minute give up and die!
            while(GetTotalThreadCount() > 0 && liImpatienceCounter < 30)
            {
                printf("Waiting for %d remaining threads to die..\n", liActiveThreads);
                RSLEEP(2000); //Sleep while waiting for threads to die.
                liImpatienceCounter++;
            }
        }
        printf("All threads are dead.\n");

        //Clean up rooted compiled script.
        JS_RemoveRootRT(g_pRuntime, &g_pJSScriptObj);
    }

    
    //Cleanup any contexts that we have in our free pool.
    LockPool

    JSCONTEXT_POOL_LIST::iterator oCxIter;
    JSContext* pContext=NULL;
    oCxIter = g_oFreeContextsList.begin();
    while(oCxIter != g_oFreeContextsList.end())
    {
        pContext = (*oCxIter);
        JS_SetContextThread(pContext);          //Set to current thread so destroy doesn't assert.
        JS_SetGlobalObject(pContext, NULL);		//Make sure globals are freed too.
		printf("STARTING GARBAGE COLLECTION!\n");
        JS_GC(pContext);	                    //Gargage collect so all object finalizers are called.
        JS_DestroyContext(pContext);
        ++oCxIter;
    }
    g_oFreeContextsList.clear();

    UnlockPool


	JS_DestroyRuntime(g_pRuntime);
	JS_ShutDown();	//Free anything left hanging.
    
#ifdef WIN32
    DeleteCriticalSection(g_pContextPoolLock);
#else
    pthread_mutex_destroy(g_pContextPoolLock);
#endif

    g_pContextPoolLock=NULL;
	return 0;
}
